@building-option +create-set-co
@create binary-mode http handler named generic XMPP s2s handle:generic XMPP s2s handle
@prop _."stream" #-1 rc
@prop _."type_handlers" {} ""
@prop _."outbound" 0 c
@prop _."output_lines" {} ""
@prop _."timeout_task" 0 rc
@prop _."queued_lines" {} rc
@prop _."debug" 0 rc
;_.("error_dumps") = {#2}
;_.("daemon") = #5269
;_.("aliases") = {"generic XMPP s2s handle"}

@verb _:"read" this none this
@program _:read
if (caller != this)
  raise(E_PERM);
endif
if (this.queued_lines)
  l = this.queued_lines[1];
  this.queued_lines = listdelete(this.queued_lines, 1);
  return l;
else
  msg = "";
  while (!msg)
    try
      msg = read(this);
    except v (ANY)
      this:abort();
      kill_task(task_id());
    endtry
    {full, partial} = this:next_tag(tostr(this.partial_line, msg));
    this.partial_line = partial;
    if (length(full) > 1)
      this.queued_lines = listdelete(full, 1);
      msg = full[1];
    elseif (length(full) == 1)
      msg = full[1];
    else
      msg = "";
    endif
  endwhile
  this.debug && this.owner:debug({full, partial}, tostr(this, " recv"));
  return msg;
endif
.

@verb _:"connected" this none this
@program _:connected
if (caller in {this, this.daemon})
  if (this.timeout)
    fork timeout_task (this.timeout)
      `this:check_timeout() ! ANY';
    endfork
    this.timeout_task = timeout_task;
  endif
endif
fork t (0)
  try
    this:set_busy(task_id());
    if ($string_utils:connection_direction(this:connection_name()) == "to")
      this.outbound = 1;
    endif
    this:handle_command();
    this:set_busy(0);
  except v (ANY)
    this.owner:debug(v);
    v = this:remap_traceback(v);
    this:report_traceback(v);
    this:respond_server_error(v);
    this:close();
  endtry
endfork
this:set_busy(t);
.

@verb _:"origin_host" this none this
@program _:origin_host
if (caller == this)
  return $string_utils:connection_hostname(this:connection_name());
else
  raise(E_PERM);
endif
.

@verb _:"handle_command" this none this
@program _:handle_command
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
this.stream = $local.jabber:new_type_handler("stream");
this.stream.connection = this;
if (this.outbound)
  origin_host = $string_utils:connection_hostname(this:connection_name());
  {fromhost, tohost} = $local.jabber:find_outbound_tofrom(origin_host);
  this.stream.from = fromhost;
  this.stream.to = tohost;
  this.stream:set_connected(1);
  this:send(this.stream);
  input = this:read();
  stream_data = $local.xml:parse(tostr(input, "</stream:stream>"))[4][1][2];
  this.stream:set_many(@stream_data);
  dbr = this:get_type_handler("db:result");
  dbr.from = this.stream.from;
  dbr.to = this.stream.to;
  dbr.text = string_hash(tostr(time(), this, dbr.to));
  this:send(dbr);
else
  input = this:read();
  if ((!index(input, "<?xml")) && (!index(input, "<stream:stream")))
    raise(E_INVIND, "Invalid Stream Header!");
  endif
  stream_data = $local.xml:parse(tostr(input, "</stream:stream>"))[4][1][2];
  this.stream:set_many(@stream_data);
  this.stream:set_connected(1);
  if (this.stream.xmlns == "jabber:server")
    this:send(this.stream);
  endif
endif
while (this.stream:connected())
  xml_input = this:read();
  input = $local.xml:parse(xml_input)[4][1];
  handler = this:get_type_handler(input[1]);
  if ($waif:isa(handler, $waif))
    handler:handle(this, @input);
  endif
  if (this.stream:connected(2))
    while (this.output_lines)
      msg = this.output_lines[1];
      this.output_lines = listdelete(this.output_lines, 1);
      this:notify(msg);
    endwhile
  endif
endwhile
this:close();
.

@verb _:"set_many" this none this
@program _:set_many
if (!$dict:is_dict(this.dict))
  this.dict = $dict.empty;
endif
for field in (args)
  {key, data} = field;
  this.dict[key] = data;
endfor
.

@verb _:"do_command" this none this
@program _:do_command
this.owner:tell(argstr);
return 1;
.

@verb _:"send" this none this
@program _:send
{object} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
msg = object:serialize();
if (`this.stream.established ! ANY => 0')
  return this:notify(msg);
else
  return this:enqueue_output(msg);
endif
.

@verb _:"get_type_handler" this none this
@program _:get_type_handler
{type} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
subtype = "";
if ((delim = index(type, ":")) > 1)
  subtype = type[delim + 1..$];
  type = type[1..delim - 1];
endif
for handler in (this.type_handlers)
  if ((handler.class.stanza == type) && `handler.subtype == subtype ! E_VARNF, E_PROPNF => 1')
    return handler;
  endif
endfor
if ($waif:isa(new_handler = $local.jabber:new_type_handler(type), $waif))
  this.type_handlers = {@this.type_handlers, new_handler};
  `new_handler.connection = this ! E_PROPNF';
  subtype && `new_handler.subtype = subtype ! E_PROPNF';
  return new_handler;
endif
return $nothing;
.

@verb _:"enqueue_input" this none this
@program _:enqueue_input
{input} = args;
if (caller in {#0})
  this.input_lines = {@this.input_lines, input};
endif
.

@verb _:"notify" this none this
@program _:notify
this.debug && this.owner:debug(args, tostr(this, " xmit"));
return pass(@args);
.

@verb _:"enqueue_output" this none this
@program _:enqueue_output
{input} = args;
if (caller in {this})
  this.output_lines = {@this.output_lines, input};
endif
.

@verb _:"next_tag" this none this
@program _:next_tag
{input} = args;
working = "";
full = {};
tagdepth = 0;
while ((length(input) > 0) && (input[1] == "~"))
  input = input[4..$];
endwhile
if ((length($string_utils:triml(input)) > 0) && (!index(input, "<")))
  raise(E_INVARG, tostr("Invalid stream format: ", input));
endif
while (idx = index(input, "<"))
  if (ridx = index(input, ">"))
    if (input[idx + 1] == "/")
      tagdepth = tagdepth - 1;
    elseif ((input[idx + 1] == "?") || (input[ridx - 1] == "/"))
      "don't increment tag depth";
    elseif (`input[idx + 1..idx + 13] == "stream:stream" ! E_RANGE => 0')
      "don't count streams";
    else
      tagdepth = tagdepth + 1;
    endif
    working = tostr(working, input[1..ridx]);
    input = `input[ridx + 1..$] ! E_RANGE => ""';
  else
    "Can't find the end of the current tag and we're still deep.";
    return {"", tostr(working, input)};
  endif
  if (tagdepth == 0)
    full = {@full, working};
    working = "";
    while ((length(input) > 0) && (input[1] == "~"))
      input = input[4..$];
    endwhile
  endif
endwhile
if (`full[1][1..5] == "<?xml" ! E_RANGE => 0' && (length(full) > 1))
  first = tostr(@full[1..2]);
  full = {first, @`full[3..$] ! E_RANGE => {}'};
endif
return {full, tostr(working, input)};
.

@verb _:"remove_type_handler" this none this
@program _:remove_type_handler
{type} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
subtype = "";
if ((delim = index(type, ":")) > 1)
  subtype = type[delim + 1..$];
  type = type[1..delim - 1];
endif
for handler in (this.type_handlers)
  if ((handler.class.stanza == type) && `handler:get_subtype() == subtype ! E_VARNF => 1')
    this.handlers = setremove(this.handlers, handler);
    return 1;
  endif
endfor
return 0;
.

@verb _:"report_traceback" this none this
@program _:report_traceback
{v} = args;
if (caller != this)
  return;
endif
formatted = $code_utils:format_traceback(v[4], v[1], v[2]);
for dude in (this.error_dumps)
  dude:tell("+----JABBER TRACEBACK");
  dude:tell("| from: ", this:origin_host());
  for line in (formatted)
    dude:tell("| ", line);
  endfor
  dude:Tell("+-----");
endfor
return;
.

@verb _:"connection_name" this none this rxd #999
@program _:connection_name
if (caller == this)
  return connection_name(this);
else
  raise(E_PERM);
endif
.

@verb _:"respond_server_error" this none this
@program _:respond_server_error
{v} = args;
this:notify(tostr("<stream:error>", $local.xml:xml_escape(v[2]), "</stream:error>"));
if (`this.stream:connected() ! E_INVIND => 0')
  this:notify("</stream:stream>");
endif
$network:close(this);
return;
.

@verb _:"remap_traceback" this none this
@program _:remap_traceback
{v} = args;
if (caller != this)
  return;
endif
remapped = {};
for frame in (v[4])
  ths = frame[1];
  remapped = {@remapped, `frame[1]:remap_traceback_frame(frame) ! ANY => frame'};
endfor
return listset(v, remapped, 4);
.

@verb _:"disconnected" this none this
@program _:disconnected
"Usage:  :disconnected()";
"";
if (caller in {this, this.daemon})
  `kill_task(this.timeout_task) ! ANY';
  `kill_task(this.busy) ! ANY';
endif
.

@verb _:"initialize" this none this
@program _:initialize
"Usage:  :initialize()";
"";
if ((caller == this) || $perm_utils:controls(caller_perms(), this))
  ret = pass(@args);
  this:set_name(strsub(parent(this).name, "generic ", ""));
  return ret;
else
  raise(E_PERM);
endif
"Copied from generic daemon service handler (#1807):initialize by Kenny (#267) Mon Oct 10 08:35:30 2005 PDT";
.

"***finished***
